# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import time
import Queue
from unittest import TestCase

from mock import Mock, patch, call
from fixture import TempIO

from rattail.config import make_config
from rattail.filemon import actions
from rattail.filemon.config import Profile, ProfileAction


class TestAction(TestCase):

    def test_callable_must_be_implemented_in_subclass(self):
        config = make_config([])
        action = actions.Action(config)
        self.assertRaises(NotImplementedError, action)


@patch(u'rattail.filemon.actions.noop')
class TestPerformActions(TestCase):

    def setUp(self):
        self.tmp = TempIO()
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'monitor', u'foo')
        self.config.set(u'rattail.filemon', u'foo.dirs', self.tmp)
        self.config.set(u'rattail.filemon', u'foo.actions', u'noop')
        self.config.set(u'rattail.filemon', u'foo.action.noop.func', u'rattail.filemon.actions:noop')
        # Must delay creating the profile since doing it here would bypass our mock of noop.

    def get_profile(self, stop_on_error=False):
        profile = Profile(self.config, u'foo')
        profile.stop_on_error = stop_on_error
        profile.queue = Mock()
        profile.queue.get_nowait.side_effect = [
            Queue.Empty, # for coverage sake; will be effectively skipped
            self.tmp.putfile(u'file1', u''),
            self.tmp.putfile(u'file2', u''),
            self.tmp.putfile(u'file3', u''),
            actions.StopProcessing,
            ]
        return profile

    def test_action_is_invoked_for_each_file_in_queue(self, noop):
        profile = self.get_profile()
        actions.perform_actions(profile)
        self.assertEqual(noop.call_count, 3)
        noop.assert_has_calls([
                call(self.tmp.join(u'file1')),
                call(self.tmp.join(u'file2')),
                call(self.tmp.join(u'file3')),
                ])

    def test_action_is_skipped_for_nonexistent_file(self, noop):
        profile = self.get_profile()
        os.remove(self.tmp.join(u'file2'))
        actions.perform_actions(profile)
        self.assertEqual(noop.call_count, 2)
        # no call for file2
        noop.assert_has_calls([
                call(self.tmp.join(u'file1')),
                call(self.tmp.join(u'file3')),
                ])

    def test_action_which_raises_error_causes_subsequent_actions_to_be_skipped_for_same_file(self, noop):
        self.config.set(u'rattail.filemon', u'foo.actions', u'noop, delete')
        self.config.set(u'rattail.filemon', u'foo.action.delete.func', u'os:remove')
        profile = self.get_profile()
        # processing second file fails, so it shouldn't be deleted
        noop.side_effect = [None, RuntimeError, None]
        actions.perform_actions(profile)
        self.assertFalse(os.path.exists(self.tmp.join(u'file1')))
        self.assertTrue(os.path.exists(self.tmp.join(u'file2')))
        self.assertFalse(os.path.exists(self.tmp.join(u'file3')))

    def test_action_which_raises_error_causes_all_processing_to_stop_if_so_configured(self, noop):
        self.config.set(u'rattail.filemon', u'foo.actions', u'noop, delete')
        self.config.set(u'rattail.filemon', u'foo.action.delete.func', u'os:remove')
        profile = self.get_profile(stop_on_error=True)
        # processing second file fails; third file shouldn't be processed at all
        noop.side_effect = [None, RuntimeError, None]
        actions.perform_actions(profile)
        self.assertEqual(noop.call_count, 2)
        noop.assert_has_calls([
                call(self.tmp.join(u'file1')),
                call(self.tmp.join(u'file2')),
                ])
        self.assertFalse(os.path.exists(self.tmp.join(u'file1')))
        self.assertTrue(os.path.exists(self.tmp.join(u'file2')))
        self.assertTrue(os.path.exists(self.tmp.join(u'file3')))


class TestInvokeAction(TestCase):

    def setUp(self):
        self.action = ProfileAction()
        self.action.action = Mock(return_value=None)
        self.action.retry_attempts = 6
        self.tmp = TempIO()
        self.file = self.tmp.putfile(u'file', u'')

    def test_action_which_succeeds_is_only_called_once(self):
        actions.invoke_action(self.action, self.file)
        self.assertEqual(self.action.action.call_count, 1)

    def test_action_with_no_delay_does_not_pause_between_attempts(self):
        self.action.retry_attempts = 3
        self.action.action.side_effect = [RuntimeError, RuntimeError, None]
        start = time.time()
        actions.invoke_action(self.action, self.file)
        self.assertEqual(self.action.action.call_count, 3)
        self.assertTrue(time.time() - start < 1.0)

    def test_action_with_delay_pauses_between_attempts(self):
        self.action.retry_attempts = 3
        self.action.retry_delay = 1
        self.action.action.side_effect = [RuntimeError, RuntimeError, None]
        start = time.time()
        actions.invoke_action(self.action, self.file)
        self.assertEqual(self.action.action.call_count, 3)
        self.assertTrue(time.time() - start >= 2.0)

    def test_action_which_fails_is_only_attempted_the_specified_number_of_times(self):
        self.action.action.side_effect = RuntimeError
        # Last attempt will not handle the exception; assert that as well.
        self.assertRaises(RuntimeError, actions.invoke_action, self.action, self.file)
        self.assertEqual(self.action.action.call_count, 6)

    def test_action_which_fails_then_succeeds_stops_retrying(self):
        # First 2 attempts fail, third succeeds.
        self.action.action.side_effect = [RuntimeError, RuntimeError, None]
        actions.invoke_action(self.action, self.file)
        self.assertEqual(self.action.action.call_count, 3)

    def test_action_which_fails_with_different_errors_stops_retrying(self):
        self.action.action.side_effect = [ValueError, TypeError, None]
        # Second attempt will not handle the exception; assert that as well.
        self.assertRaises(TypeError, actions.invoke_action, self.action, self.file)
        self.assertEqual(self.action.action.call_count, 2)


class TestRaiseException(TestCase):

    def test_exception_is_raised(self):
        # this hardly deserves a test, but what the hell
        self.assertRaises(Exception, actions.raise_exception, '/dev/null')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync for Linux
"""

from __future__ import unicode_literals
from __future__ import absolute_import

import time
import datetime
import logging

from rattail.db import api
from rattail.daemon import Daemon
from rattail.threads import Thread
from rattail.datasync.config import load_profiles


log = logging.getLogger(__name__)


class DataSyncDaemon(Daemon):
    """
    Linux daemon implementation of DataSync.
    """

    def run(self):
        """
        Starts watcher and consumer threads according to configuration.
        """
        for key, profile in load_profiles(self.config).iteritems():

            # Create watcher thread for the profile.
            name = '{0}-watcher'.format(key)
            log.debug("starting thread '{0}' with watcher: {1}".format(name, profile.watcher_spec))
            thread = Thread(target=watch_for_changes, name=name, args=(self.config, profile.watcher))
            thread.daemon = True
            thread.start()

            # Create consumer threads, unless watcher consumes itself.
            if not profile.watcher.consumes_self:

                # Create a thread for each "isolated" consumer.
                for consumer in profile.isolated_consumers:
                    name = '{0}-consumer-{1}'.format(key, consumer.key)
                    log.debug("starting thread '{0}' with isolated consumer: {1}".format(name, consumer.spec))
                    thread = Thread(target=consume_changes, name=name, args=(profile, [consumer]))
                    thread.daemon = True
                    thread.start()

                # Maybe create a common (shared transaction) thread for the rest of
                # the consumers.
                if profile.common_consumers:
                    name = '{0}-consumer-shared'.format(key)
                    log.debug("starting thread '{0}' with consumer(s): {1}".format(
                        name, ','.join(["{0} ({1})".format(c.key, c.spec) for c in profile.common_consumers])))
                    thread = Thread(target=consume_changes, name=name, args=(profile, profile.common_consumers,))
                    thread.daemon = True
                    thread.start()

        # Loop indefinitely.  Since this is the main thread, the app will
        # terminate when this method ends; all other threads are "subservient"
        # to this one.
        while True:
            time.sleep(.01)


def watch_for_changes(config, watcher):
    """
    Target for DataSync watcher threads.
    """
    from rattail.db import Session, model

    # Let watcher do any setup it needs.
    watcher.setup()

    # The 'last run' value is maintained as a naive UTC datetime.
    timefmt = '%Y-%m-%d %H:%M:%S'
    lastrun_setting = 'rattail.datasync.{0}.watcher.lastrun'.format(watcher.key)
    lastrun = api.get_setting(None, lastrun_setting)
    if lastrun:
        lastrun = datetime.datetime.strptime(lastrun, timefmt)

    error = False
    while not error:

        thisrun = datetime.datetime.utcnow()

        try:
            changes = watcher.get_changes(lastrun)
        except:
            log.exception("error calling watcher.get_changes(), this watcher will "
                          "now *terminate* (until datasync is restarted)")
            error = True
        else:

            lastrun = thisrun
            api.save_setting(None, lastrun_setting, lastrun.strftime(timefmt))
            if changes:
                log.debug("got {0} changes from '{1}' watcher".format(len(changes), watcher.key))

                # If watcher consumes itself, then it will process its own
                # changes.  Note that there are no assumptions made about the
                # format or structure of these change objects.
                if watcher.consumes_self:
                    session = Session()
                    try:
                        watcher.process_changes(session, changes)
                    except Exception:
                        log.exception("error calling watcher.process_changes()")
                        session.rollback()
                    else:
                        session.commit()
                        log.debug("watcher has consumed its own changes")
                    finally:
                        session.close()

                else:
                    # Give all change stubs the same timestamp, to help identify them
                    # as a "batch" of sorts, so consumers can process them as such.
                    now = datetime.datetime.utcnow()

                    # Save change stub records to rattail database, for consumer thread
                    # to find and process.
                    saved = 0
                    session = Session()
                    for key, change in changes:
                        for consumer in watcher.consumer_stub_keys:
                            session.add(model.DataSyncChange(
                                source=watcher.key,
                                payload_type=change.payload_type,
                                payload_key=change.payload_key,
                                deletion=change.deletion,
                                obtained=now,
                                consumer=consumer))
                            saved += 1
                        session.flush()
                    session.commit()
                    session.close()
                    log.debug("saved {0} '{1}' change stubs to rattail database".format(saved, watcher.key))

                    # Tell watcher to prune the original change records from its source
                    # database, if relevant.  Note that we only give it the keys for this.
                    try:
                        pruned = watcher.prune_changes([c[0] for c in changes])
                    except:
                        log.exception("error calling watcher.prune_changes()")
                    else:
                        if pruned is not None:
                            log.debug("pruned {0} changes from '{1}' database".format(pruned, watcher.key))

        time.sleep(watcher.delay)


def consume_changes(profile, consumers):
    """
    Target for DataSync consumer thread.
    """
    from rattail.db import Session, model

    # Let consumers do any setup they need.
    for consumer in consumers:
        consumer.setup()

    def process(session, consumer, changes):
        try:
            consumer.process_changes(session, changes)
        except:
            log.exception("error calling consumer.process_changes(), this consumer "
                          "will now *terminate* (until datasync is restarted)")
            return False
        for change in changes:
            session.delete(change)
            session.flush()
        return True

    error = False
    while True:

        session = Session()
        for consumer in consumers:

            changes = session.query(model.DataSyncChange)\
                             .filter(model.DataSyncChange.source == profile.key)\
                             .filter(model.DataSyncChange.consumer == consumer.key)\
                             .order_by(model.DataSyncChange.obtained)\
                             .all()
                
            if changes:
                log.debug("found {} changes to process".format(len(changes)))
                consumer.begin_transaction()

                # Process (consume) changes in batches, according to timestamp.
                batch = []
                batchcount = 1
                changecount = 0
                timestamp = None
                for change in changes:
                    if timestamp and change.obtained != timestamp:
                        if process(session, consumer, batch):
                            changecount += len(batch)
                        else:
                            error = True
                            break
                        batch = []
                        batchcount += 1
                    batch.append(change)
                    timestamp = change.obtained
                if not error:
                    if process(session, consumer, batch):
                        changecount += len(batch)
                    else:
                        error = True

                log.debug("processed {} changes in {} batches{}".format(
                    changecount, batchcount, " (with errors)" if error else ""))
                if error:
                    consumer.rollback_transaction()
                    break
                else:
                    consumer.commit_transaction()

        if error:
            session.rollback()
            session.close()
            break
        else:
            session.commit()
            session.close()

        time.sleep(profile.consumer_delay)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Stores
"""

from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list

from .core import Base, uuid_column
from .contact import PhoneNumber, EmailAddress


class Store(Base):
    """
    Represents a store (physical or otherwise) within the organization.
    """
    __tablename__ = 'store'
    __versioned__ = {}

    uuid = uuid_column()
    id = sa.Column(sa.String(length=10))
    name = sa.Column(sa.String(length=100))
    database_key = sa.Column(sa.String(length=30))

    def __unicode__(self):
        return unicode(self.name or '')

    def add_email_address(self, address, type='Info'):
        email = StoreEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Voice'):
        phone = StorePhoneNumber(number=number, type=type)
        self.phones.append(phone)


class StorePhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a store.
    """

    __mapper_args__ = {'polymorphic_identity': 'Store'}


Store.phones = relationship(
    StorePhoneNumber,
    backref='store',
    primaryjoin=StorePhoneNumber.parent_uuid == Store.uuid,
    foreign_keys=[StorePhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=StorePhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Store.phone = relationship(
    StorePhoneNumber,
    primaryjoin=sa.and_(
        StorePhoneNumber.parent_uuid == Store.uuid,
        StorePhoneNumber.preference == 1),
    foreign_keys=[StorePhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class StoreEmailAddress(EmailAddress):
    """
    Represents an email address associated with a store.
    """

    __mapper_args__ = {'polymorphic_identity': 'Store'}


Store.emails = relationship(
    StoreEmailAddress,
    backref='store',
    primaryjoin=StoreEmailAddress.parent_uuid == Store.uuid,
    foreign_keys=[StoreEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=StoreEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Store.email = relationship(
    StoreEmailAddress,
    primaryjoin=sa.and_(
        StoreEmailAddress.parent_uuid == Store.uuid,
        StoreEmailAddress.preference == 1),
    foreign_keys=[StoreEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)
